import sys
import atexit


MPI = None

def _import_mpi(use_dill=False):
    global MPI
    try:
        from mpi4py import MPI as _MPI
        if use_dill:
            import dill
            _MPI.pickle.__init__(dill.dumps, dill.loads, dill.HIGHEST_PROTOCOL)
        MPI = _MPI
    except:
        raise ImportError("Please install mpi4py")

    return MPI


class MPIPool:
    """A processing pool that distributes tasks using MPI.
    With this pool class, the master process distributes tasks to worker
    processes using an MPI communicator.
    This implementation is inspired by @juliohm in `this module
    <https://github.com/juliohm/HUM/blob/master/pyhum/utils.py#L24>`_
    and was adapted from schwimmbad.
    Parameters
    ----------
    comm : :class:`mpi4py.MPI.Comm`, optional
        An MPI communicator to distribute tasks with. If ``None``, this uses
        ``MPI.COMM_WORLD`` by default.
    """

    def __init__(self, comm=None):

        self.comm = MPI.COMM_WORLD if comm is None else comm

        self.master = 0
        self.rank = self.comm.Get_rank()

        atexit.register(lambda: MPIPool.close(self))

        if not self.is_master():
            # workers branch here and wait for work
            self.wait()
            sys.exit(0)

        self.workers = set(range(self.comm.size))
        self.workers.discard(self.master)
        self.size = self.comm.Get_size() - 1

        if self.size == 0:
            raise ValueError("Tried to create an MPI pool, but there "
                             "was only one MPI process available. "
                             "Need at least two.")


    def wait(self):
        """Tell the workers to wait and listen for the master process. This is
        called automatically when using :meth:`MPIPool.map` and doesn't need to
        be called by the user.
        """
        if self.is_master():
            return

        status = MPI.Status()
        while True:
            task = self.comm.recv(source=self.master, tag=MPI.ANY_TAG, status=status)

            if task is None:
                # Worker told to quit work
                break

            func, arg = task
            result = func(arg)
            # Worker is sending answer with tag
            self.comm.ssend(result, self.master, status.tag)


    def map(self, worker, tasks):
        """Evaluate a function or callable on each task in parallel using MPI.
        The callable, ``worker``, is called on each element of the ``tasks``
        iterable. The results are returned in the expected order.
        Parameters
        ----------
        worker : callable
            A function or callable object that is executed on each element of
            the specified ``tasks`` iterable. This object must be picklable
            (i.e. it can't be a function scoped within a function or a
            ``lambda`` function). This should accept a single positional
            argument and return a single object.
        tasks : iterable
            A list or iterable of tasks. Each task can be itself an iterable
            (e.g., tuple) of values or data to pass in to the worker function.

        Returns
        -------
        results : list
            A list of results from the output of each ``worker()`` call.
        """

        # If not the master just wait for instructions.
        if not self.is_master():
            self.wait()
            return


        workerset = self.workers.copy()
        tasklist = [(tid, (worker, arg)) for tid, arg in enumerate(tasks)]
        resultlist = [None] * len(tasklist)
        pending = len(tasklist)

        while pending:
            if workerset and tasklist:
                worker = workerset.pop()
                taskid, task = tasklist.pop()
                # "Sent task %s to worker %s with tag %s"
                self.comm.send(task, dest=worker, tag=taskid)

            if tasklist:
                flag = self.comm.Iprobe(source=MPI.ANY_SOURCE, tag=MPI.ANY_TAG)
                if not flag:
                    continue
            else:
                self.comm.Probe(source=MPI.ANY_SOURCE, tag=MPI.ANY_TAG)

            status = MPI.Status()
            result = self.comm.recv(source=MPI.ANY_SOURCE, tag=MPI.ANY_TAG,
                                    status=status)
            worker = status.source
            taskid = status.tag

            # "Master received from worker %s with tag %s"

            workerset.add(worker)
            resultlist[taskid] = result
            pending -= 1

        return resultlist


    def close(self):
        """ Tell all the workers to quit."""
        if self.is_worker():
            return

        for worker in self.workers:
            self.comm.send(None, worker, 0)


    def is_master(self):
        return self.rank == 0


    def is_worker(self):
        return self.rank != 0


    def __enter__(self):
        return self


    def __exit__(self, *args):
        self.close()



def split_ranks(N_ranks, N_chunks):
    """
    Divide the ranks into N chunks
    """
    seq = range(N_ranks)
    avg = int(N_ranks // N_chunks)
    remainder = N_ranks % N_chunks

    start = 0
    end = avg
    for i in range(N_chunks):
        if remainder:
            end += 1
            remainder -= 1
        yield i, seq[start:end]
        start = end
        end += avg



class ChainManager:
    """
    Class to serve as context manager to handle to MPI-related issues,
    specifically, the managing of ``MPIPool`` and splitting of communicators.
    This class can be used to run ``nchains`` in parallel with each chain
    having its own ``MPIPool`` of parallel walkers.
    
    Parameters
    ----------
    nchains : int
        the number of independent chains to run concurrently
    comm : MPI.Communicator
        the global communicator to split
    """

    def __init__(self, nchains=1, comm=None):
        global MPI
        MPI = _import_mpi(use_dill=False)

        self.comm  = MPI.COMM_WORLD if comm is None else comm
        self.nchains = nchains

        # initialize comm for parallel chains
        self.chains_group = None
        self.chains_comm  = None

        # intiialize comm for pool of workers for each parallel chain
        self.pool_comm = None
        self.pool      = None


    def __enter__(self):
        """
        Setup the MPIPool, such that only the ``pool`` master returns,
        while the other processes wait for tasks
        """
        # split ranks if we need to
        if self.comm.size > 1:

            ranges = []
            for i, ranks in split_ranks(self.comm.size, self.nchains):
                ranges.append(ranks[0])
                if self.comm.rank in ranks:
                    color = i

            # split the global comm into pools of workers
            self.pool_comm = self.comm.Split(color, 0)

            # make the comm to communicate b/w parallel runs
            if self.nchains >= 1:
                self.chains_group = self.comm.group.Incl(ranges)
                self.chains_comm = self.comm.Create(self.chains_group)

        # initialize the MPI pool, if the comm has more than 1 process
        if self.pool_comm is not None and self.pool_comm.size > 1:
            self.pool = MPIPool(comm=self.pool_comm)

        # explicitly force non-master ranks in pool to wait
        if self.pool is not None and not self.pool.is_master():
            self.pool.wait()
            sys.exit(0)

        self.rank = 0
        if self.chains_comm is not None:
            self.rank = self.chains_comm.rank

        return self


    def __exit__(self, exc_type, exc_value, exc_traceback):
        """
        Exit gracefully by closing and freeing the MPI-related variables
        """
        # wait for all the processes, if we more than one
        if self.chains_comm is not None and self.chains_comm.size > 1:
            self.chains_comm.Barrier()

        # close and free the MPI stuff
        if self.chains_group is not None:
            self.chains_group.Free()
        if self.chains_comm is not None:
            self.chains_comm.Free()
        if self.pool is not None:
            self.pool.close()

        return True


    @property
    def get_rank(self):
        '''
        Get ``rank`` of current ``chain``. The minimum ``rank`` is ``0`` and the maximum is ``nchains-1``.
        '''
        return self.rank


    @property
    def get_pool(self):
        '''
        Get parallel ``pool`` of workers that correspond to a specific chain. This should be used to
        parallelize the walkers of each ``chain`` (not the chains themselves). This includes the ``map``
        method that ``zeus`` requires.
        '''
        return self.pool

    
    def gather(self, x, root):
        '''
        Gather method to gather ``x`` in ``rank = root`` chain.

        Parameters
        ----------
        x : Python object
            The python object to be gathered.
        root : int
            The rank  of the chain that x is gathered.

        Returns
        -------
        x : Python object
            The input object x gathered in ``rank = root``.
        '''
        return self.chains_comm.gather(x, root=root)

    
    def scatter(self, x, root):
        '''
        Scatter method to scatter ``x`` from ``rank = root`` chain to the rest. 

        Parameters
        ----------
        x : Python object
            The python object to be scattered.
        root : int
            The rank of the origin chain from which the x is scattered.

        Returns
        -------
        x : Pythonn object
            Part of the input object x that was scattered along the ranks.
        '''
        return self.chains_comm.scatter(x, root=root)


    def allgather(self, x):
        '''
        Allgather method to gather ``x`` in all chains. This is equivalent to first ``scatter`` and then ``bcast``.

        Parameters
        ----------
        x : Python object
            The python object to be gathered.

        Returns
        -------
        x : Python object
            The python object, gathered in all ranks.
        '''
        return self.chains_comm.allgather(x)


    def bcast(self, x, root):
        '''
        Broadcast method to send ``x`` from ``rank = root`` to all chains.

        Parameters
        ----------
        x : Python object
            The python object to be send.
        root : int
            The rank of the origin chain from which the object x is sent.
        
        Returns
        -------
        x : Python object
            The input object x in all ranks.
        '''
        return self.chains_comm.bcast(x, root=root)


import numpy as np
from scipy.fft import fft, ifft

def _autocorr_func_1d(x, norm=True):
    """
    Autocorrelation Function of 1-dimensional chain.

    Args:
        x (array) : 1-dimensional chain.
        norm (bool) : By default norm=True and the autocorrelation function will be normalized.

    Returns:
        The (normalised if norm=True) autocorrelation function of the chain x.
    """
    x = np.atleast_1d(x)
    if len(x.shape) != 1:
        raise ValueError("invalid dimensions for 1D autocorrelation function")

    # Next largest power of 2
    n = 1
    while n < len(x):
        n = n << 1

    # Compute the auto-correlation function using FFT
    #f = np.fft.fft(x - np.mean(x), n=2 * n)
    #acf = np.fft.ifft(f * np.conjugate(f))[: len(x)].real
    f = fft(x - np.mean(x), n=2 * n)
    acf = ifft(f * np.conjugate(f))[: len(x)].real
    acf /= 4 * n

    # Normalize
    if norm:
        acf /= acf[0]

    return acf


def _autocorr_time_1d(y, c=5.0, method='mk'):
    """
    Integrated Autocorrelation Time (IAT) for 1-dimensional chain.

    Args:
        y (array) : (nsteps, nwalkers) array for one parameter.
        c (float) : Truncation parameter of automated windowing procedure of Sokal (1989), default is 5.0.
        method (str) : Method to use to compute the IAT. Available options are ``mk`` (Default), ``dfm``, and ``gw``.

    Returns:
        The IAT of the chain y.
    """

    if method not in ['mk', 'dfm', 'gw']:
        raise ValueError('Please select one of the supported methods i.e. mk (Recommended), dfm, gw.')

    if method == 'mk':
        # Minas Karamanis method
        f = _autocorr_func_1d(y.reshape((-1), order='C'))
    elif method == 'dfm':
        # Daniel Forman-Mackey method
        f = np.zeros(y.shape[1])
        for yy in y:
            f += _autocorr_func_1d(yy)
        f /= len(y)
    else:
        # Goodman-Weary method
        f = _autocorr_func_1d(np.mean(y, axis=0))
    
    taus = 2.0 * np.cumsum(f) - 1.0

    # Automated windowing procedure following Sokal (1989)
    def auto_window(taus, c):
        m = np.arange(len(taus)) < c * taus
        if np.any(m):
            return np.argmin(m)
        return len(taus) - 1

    window = auto_window(taus, c)
    return taus[window]


def AutoCorrTime(samples, c=5.0, method='mk'):
    """
    Integrated Autocorrelation Time (IAT) for all the chains.

    Parameters
    ----------
        samples : array
            3-dimensional array of shape (nsteps, nwalkers, ndim)
        c : float
            Truncation parameter of automated windowing procedure of Sokal (1989), default is 5.0
        method : str
            Method to use to compute the IAT. Available options are ``mk`` (Default), ``dfm``, and ``gw``.

    Returns
    -------
    taus : array
        Array with the IAT of all the chains.
    """

    if method not in ['mk', 'dfm', 'gw']:
        raise ValueError('Please select one of the supported methods i.e. mk (Recommended), dfm, gw.')

    _, _, ndim = np.shape(samples)

    taus = np.empty(ndim)
    for i in range(ndim):
        taus[i] = _autocorr_time_1d(samples[:,:,i].T, c, method)

    return taus



import numpy as np

def cornerplot(samples,
               labels=None,
               weights=None,
               levels=None,
               span=None,
               quantiles=[0.025, 0.5, 0.975],
               truth=None,
               color=None,
               alpha=0.5,
               linewidth=1.5,
               fill=True,
               fontsize=10,
               show_titles=True,
               title_fmt='.2f',
               title_fontsize=12,
               cut=3,
               fig=None,
               size=(10,10)):
    r"""
    Plot corner-plot of samples.

    Parameters
    ----------
    samples : array
        Array of shape (nsamples, ndim) containing the samples.
    labels : list
        List of names of for the parameters.
    weights : array
        Array with weights (useful if different samples have different weights e.g. as in Nested Sampling).
    levels : list
        The quantiles used for plotting the smoothed 2-D distributions. If not provided, these default to 0.5, 1, 1.5, and 2-sigma contours.
    quantiles : list
        A list of fractional quantiles to overplot on the 1-D marginalized posteriors as titles. Default is ``[0.025, 0.5, 0.975]`` (spanning the 95%/2-sigma credible interval).
    truth : array
        Array specifying a point to be highlighted in the plot. It can be the true values of the parameters, the mean, median etc. By default this is None.
    color : str
        Matplotlib color to be used in the plot.
    alpha : float
        Transparency value of figure (Default is 0.5).
    linewidth : float
        Linewidth of plot (Default is 1.5).
    fill : bool
        If True (Default) the fill the 1D and 2D contours with color.
    fontsize : float
        Fontsize of axes labels. Default is 10.
    show_titles : bool
        Whether to display a title above each 1-D marginalized posterior showing the quantiles. Default is True.
    title_fmt : str
        Format of the titles. Default is ``.2f``.
    title_fontsize : float
        Fontsize of titles. Default is 12.
    cut : float
        Factor, multiplied by the smoothing bandwidth, that determines how far the evaluation grid extends past the extreme datapoints.
        When set to 0, truncate the curve at the data limits. Default is ``cut=3``.
    fig : (figure, axes)
        Pre-existing Figure and Axes for the plot. Otherwise create new internally. Default is None.
    size : (int, int)
        Size of the plot. Default is (10, 10).
    
    Returns
    -------
    Figure, Axes
        The matplotlib figure and axes.
    """
    import matplotlib.pyplot as plt
    import seaborn as sns

    ndim = samples.shape[1]

    if labels is None:
        labels = [r"$x_{"+str(i+1)+"}$" for i in range(ndim)]

    if levels is None:
        levels = list(1.0 - np.exp(-0.5 * np.arange(0.5, 2.1, 0.5) ** 2))
    levels.append(1.0)
    
    if color is None:
        color = "tab:blue"
    
    # Determine plotting bounds.
    if span is None:
        span = [0.999999426697 for i in range(ndim)]
    span = list(span)
    if len(span) != ndim:
        raise ValueError("Dimension mismatch between samples and span.")
    for i, _ in enumerate(span):
        try:
            xmin, xmax = span[i]
        except:
            q = [0.5 - 0.5 * span[i], 0.5 + 0.5 * span[i]]
            span[i] = _quantile(samples[:,i], q, weights=weights)

    idxs = np.arange(ndim**2).reshape(ndim, ndim)
    tril = np.tril_indices(ndim)
    triu = np.triu_indices(ndim)
    lower = list(set(idxs[tril])-set(idxs[triu]))
    upper = list(set(idxs[triu])-set(idxs[tril]))
    
    if fig is None:
        figure, axes = plt.subplots(ndim, ndim, figsize=size, sharex=False)
    else:
        figure = fig[0]
        axes = fig[1]

    for idx, ax in enumerate(axes.flat):

        i = idx // ndim
        j = idx % ndim        
        
        if idx in lower:
            
            ax.set_ylim(span[i])
            ax.yaxis.set_major_locator(plt.MaxNLocator(5))

            if fill:
                ax = sns.kdeplot(x=samples[:,j], y=samples[:,i], weights=weights,
                                fill=True, color=color,
                                clip=None, cut=cut,
                                thresh=levels[0], levels=levels,
                                ax=ax, alpha=alpha, linewidth=0.0,
                                )
            ax = sns.kdeplot(x=samples[:,j], y=samples[:,i], weights=weights,
                             fill=False, color=color,
                             clip=None, cut=cut,
                             thresh=levels[0], levels=levels,
                             ax=ax, alpha=alpha, linewidth=linewidth,
                             )

            if truth is not None:
                ax.axvline(truth[j], color='k', lw=1.0)
                ax.axhline(truth[i], color='k', lw=1.0)

            if j == 0:
                ax.set_ylabel(labels[i], fontsize=fontsize)
                [l.set_rotation(45) for l in ax.get_yticklabels()]
            else:
                ax.yaxis.set_ticklabels([])

            if i == ndim - 1:
                ax.xaxis.set_major_locator(plt.MaxNLocator(5))
                ax.set_xlabel(labels[j], fontsize=fontsize)
                [l.set_rotation(45) for l in ax.get_xticklabels()]
            else:
                ax.set_xticklabels([])

                        
        elif idx in upper:
            ax.set_axis_off()
        else:

            ax.yaxis.set_major_locator(plt.NullLocator())

            if fill:
                ax = sns.kdeplot(x=samples[:,j],
                                fill=True, color=color, weights=weights,
                                clip=None, cut=cut,
                                ax=ax, linewidth=0.0, alpha=alpha,
                                )
            ax = sns.kdeplot(x=samples[:,j],
                             fill=None, color=color, weights=weights,
                             clip=None, cut=cut,
                             ax=ax, linewidth=linewidth, alpha=alpha,
                             )

            if truth is not None:
                ax.axvline(truth[j], color='k', lw=1.0)

            if i == ndim - 1:
                ax.set_xlabel(labels[j], fontsize=fontsize)
                [l.set_rotation(45) for l in ax.get_xticklabels()]
            else:
                ax.set_xticklabels([])
            
            if show_titles:
                ql, qm, qh = _quantile(samples[:,i], quantiles, weights=weights)
                q_minus, q_plus = qm - ql, qh - qm
                fmt = "{{0:{0}}}".format(title_fmt).format
                title = r"${{{0}}}_{{-{1}}}^{{+{2}}}$"
                title = title.format(fmt(qm), fmt(q_minus), fmt(q_plus))
                title = "{0} = {1}".format(labels[i], title)
                ax.set_title(title, fontsize=title_fontsize)
            

        ax.set_xlim(span[j])
        ax.xaxis.set_major_locator(plt.MaxNLocator(5))
    
    figure.subplots_adjust(top=0.95, right=0.95, wspace=.05, hspace=.05)

    return figure, axes


def _quantile(x, q, weights=None):
    """
    Compute (weighted) quantiles from an input set of samples.
    Parameters
    ----------
    x : `~numpy.ndarray` with shape (nsamps,)
        Input samples.
    q : `~numpy.ndarray` with shape (nquantiles,)
       The list of quantiles to compute from `[0., 1.]`.
    weights : `~numpy.ndarray` with shape (nsamps,), optional
        The associated weight from each sample.
    Returns
    -------
    quantiles : `~numpy.ndarray` with shape (nquantiles,)
        The weighted sample quantiles computed at `q`.
    """

    # Initial check.
    x = np.atleast_1d(x)
    q = np.atleast_1d(q)

    # Quantile check.
    if np.any(q < 0.0) or np.any(q > 1.0):
        raise ValueError("Quantiles must be between 0. and 1.")

    if weights is None:
        # If no weights provided, this simply calls `np.percentile`.
        return np.percentile(x, list(100.0 * q))
    else:
        # If weights are provided, compute the weighted quantiles.
        weights = np.atleast_1d(weights)
        if len(x) != len(weights):
            raise ValueError("Dimension mismatch: len(weights) != len(x).")
        idx = np.argsort(x)  # sort samples
        sw = weights[idx]  # sort weights
        cdf = np.cumsum(sw)[:-1]  # compute CDF
        cdf /= cdf[-1]  # normalize CDF
        cdf = np.append(0, cdf)  # ensure proper span
        quantiles = np.interp(q, cdf, x[idx]).tolist()
        return quantiles


import numpy as np 
from itertools import permutations
import random

try:
    from scipy.stats import gaussian_kde
except ImportError:
    gaussian_kde = None

try:
    from sklearn.mixture import BayesianGaussianMixture
except ImportError:
    BayesianGaussianMixture = None



class DifferentialMove:
    r"""
    The `Karamanis & Beutler (2020) <https://arxiv.org/abs/2002.06212>`_ "Differential Move" with parallelization.
    When this Move is used the walkers move along directions defined by random pairs of walkers sampled (with no
    replacement) from the complementary ensemble. This is the default choice and performs well along a wide range
    of target distributions.

    Parameters
    ----------
        tune : bool
            If True then tune this move. Default is True.
        mu0 : float
            Default value of ``mu`` if ``tune=False``.

    """

    def __init__(self, tune=True, mu0=1.0):
        self.tune = tune
        self.mu0 = mu0


    def get_direction(self, X, mu):
        r"""
        Generate direction vectors.

        Parameters
        ----------
            X : array
                Array of shape ``(nwalkers//2, ndim)`` with the walker positions of the complementary ensemble.
            mu : float
                The value of the scale factor ``mu``.
        
        Returns
        -------
            directions : array
                Array of direction vectors of shape ``(nwalkers//2, ndim)``.
        """

        nsamples = X.shape[0]

        perms = list(permutations(np.arange(nsamples), 2))
        pairs = np.asarray(random.sample(perms,nsamples)).T

        if not self.tune:
            mu = self.mu0
        
        return 2.0 * mu * (X[pairs[0]]-X[pairs[1]]), self.tune


class GaussianMove:
    r"""
    The `Karamanis & Beutler (2020) <https://arxiv.org/abs/2002.06212>`_ "Gaussian Move" with parallelization.
    When this Move is used the walkers move along directions defined by random vectors sampled from the Gaussian
    approximation of the walkers of the complementary ensemble.

    Parameters
    ----------
        tune : bool
            If True then tune this move. Default is True.
        mu0 : float
            Default value of ``mu`` if ``tune=False``.

    """

    def __init__(self, tune=False, mu0=1.0, cov=None):
        self.tune = tune 
        self.mu0 = mu0
        self.cov = cov


    def get_direction(self, X, mu):
        r"""
        Generate direction vectors.

        Parameters
        ----------
            X : array
                Array of shape ``(nwalkers//2, ndim)`` with the walker positions of the complementary ensemble.
            mu : float
                The value of the scale factor ``mu``.
        
        Returns
        -------
            directions : array
                Array of direction vectors of shape ``(nwalkers//2, ndim)``.
        """

        nsamples = X.shape[0]

        mean = np.mean(X, axis=0)
        if self.cov is None:
            cov = np.cov(X, rowvar=False)
        else:
            cov = self.cov

        if not self.tune:
            mu = self.mu0

        return 2.0 * mu * np.random.multivariate_normal(np.zeros_like(mean),cov,size=nsamples), self.tune


class GlobalMove:
    r"""
    The `Karamanis & Beutler (2020) <https://arxiv.org/abs/2002.06212>`_ "Global Move" with parallelization.
    When this Move is used a Bayesian Gaussian Mixture (BGM) is fitted to the walkers of complementary ensemble.
    The walkers move along random directions which connect different components of the BGM in an attempt to
    facilitate mode jumping. This Move should be used when the target distribution is multimodal. This move should
    be used after any burnin period.

    Parameters
    ----------
        tune : bool
            If True then tune this move. Default is True.
        mu0 : float
            Default value of ``mu`` if ``tune=False``.
        rescale_cov : float
            Rescale the covariance matrices of the BGM components by this factor. This promotes mode jumping.
            Default value is 0.001.
        n_components : int
            The number of mixture components. Depending on the distribution of the walkers the model can
            decide not to use all of them.
    """

    def __init__(self, tune=True, mu0=1.0, rescale_cov=0.001, n_components=5):
        
        if BayesianGaussianMixture is None:
            raise ImportError("you need sklearn.mixture.BayesianGaussianMixture to use the GlobalMove")

        self.tune = tune
        self.mu0 = mu0
        self.rescale_cov = rescale_cov
        self.n_components = n_components


    def get_direction(self, X, mu):
        r"""
        Generate direction vectors.

        Parameters
        ----------
            X : array
                Array of shape ``(nwalkers//2, ndim)`` with the walker positions of the complementary ensemble.
            mu : float
                The value of the scale factor ``mu``.
        
        Returns
        -------
            directions : array
                Array of direction vectors of shape ``(nwalkers//2, ndim)``.
        """
        
        if not self.tune:
            mu = self.mu0

        n = X.shape[0]

        mixture = BayesianGaussianMixture(n_components=self.n_components)
        labels = mixture.fit_predict(X)
        means = mixture.means_
        covariances = mixture.covariances_

        i, j = np.random.choice(labels, 2, replace=False)
        if i != j:
            directions = np.random.multivariate_normal(means[i], covariances[i]*self.rescale_cov, size=n) - np.random.multivariate_normal(means[j], covariances[j]*self.rescale_cov, size=n)
            tune_once = False
        else:
            directions = mu * np.random.multivariate_normal(np.zeros_like(means[i]), covariances[i], size=n)
            if self.tune:
                tune_once = True
            else:
                tune_once = False

        return 2.0*directions, tune_once


class KDEMove:
    r"""
    The `Karamanis & Beutler (2020) <https://arxiv.org/abs/2002.06212>`_ "KDE Move" with parallelization.
    When this Move is used the distribution of the walkers of the complementary ensemble is traced using
    a Gaussian Kernel Density Estimation methods. The walkers then move along random direction vectos
    sampled from this distribution.

    Parameters
    ----------
        tune : bool
            If True then tune this move. Default is True.
        mu0 : float
            Default value of ``mu`` if ``tune=False``.
        bw_method :
            The bandwidth estimation method. See the scipy docs for allowed values.

    """

    def __init__(self, tune=False, mu0=1.0, bw_method=None):

        if gaussian_kde is None:
            raise ImportError("you need scipy.stats.gaussian_kde to use the KDEMove")

        self.tune = tune
        self.mu0 = mu0
        self.bw_method = bw_method


    def get_direction(self, X, mu):
        r"""
        Generate direction vectors.

        Parameters
        ----------
            X : array
                Array of shape ``(nwalkers//2, ndim)`` with the walker positions of the complementary ensemble.
            mu : float
                The value of the scale factor ``mu``.
        
        Returns
        -------
            directions : array
                Array of direction vectors of shape ``(nwalkers//2, ndim)``.
        """

        n = X.shape[0]

        kde = gaussian_kde(X.T, bw_method=self.bw_method)

        vectors = kde.resample(2*n).T
        directions = vectors[:n] - vectors[n:]

        if not self.tune:
            mu = self.mu0

        return 2.0 * mu * directions, self.tune


class RandomMove:
    r"""
    The `Karamanis & Beutler (2020) <https://arxiv.org/abs/2002.06212>`_ "Random Move" with parallelization.
    When this move is used the walkers move along random directions. There is no communication between the
    walkers and this Move corresponds to the vanilla Slice Sampling method. This Move should be used for
    debugging purposes only.

    Parameters
    ----------
        tune : bool
            If True then tune this move. Default is True.
        mu0 : float
            Default value of ``mu`` if ``tune=False``.

    """

    def __init__(self, tune=True, mu0=1.0):
        self.tune = tune
        self.mu0 = mu0


    def get_direction(self, X, mu):
        r"""
        Generate direction vectors.

        Parameters
        ----------
            X : array
                Array of shape ``(nwalkers//2, ndim)`` with the walker positions of the complementary ensemble.
            mu : float
                The value of the scale factor ``mu``.
        
        Returns
        -------
            directions : array
                Array of direction vectors of shape ``(nwalkers//2, ndim)``.
        """

        directions = np.random.normal(0.0, 1.0, size=X.shape)
        directions /= np.linalg.norm(directions, axis=0)

        if not self.tune:
            mu = self.mu0

        return 2.0 * mu * directions, self.tune


import numpy as np
from tqdm import tqdm
import logging

try:
    from collections.abc import Iterable
except ImportError:
    # for py2.7, will be an Exception in 3.8
    from collections import Iterable

from .samples import samples
from .fwrapper import _FunctionWrapper
from .autocorr import AutoCorrTime
from .moves import DifferentialMove


class EnsembleSampler:
    """
    An Ensemble Slice Sampler.

    Args:
        nwalkers (int): The number of walkers in the ensemble.
        ndim (int): The number of dimensions/parameters.
        logprob_fn (callable): A python function that takes a vector in the
            parameter space as input and returns the natural logarithm of the
            unnormalised posterior probability at that position.
        args (list): Extra arguments to be passed into the logp.
        kwargs (list): Extra arguments to be passed into the logp.
        moves (list): This can be a single move object, a list of moves, or a “weighted” list of the form ``[(zeus.moves.DifferentialMove(), 0.1), ...]``.
            When running, the sampler will randomly select a move from this list (optionally with weights) for each proposal. (default: DifferentialMove)
        tune (bool): Tune the scale factor to optimize performance (Default is True.)
        tolerance (float): Tuning optimization tolerance (Default is 0.05).
        patience (int): Number of tuning steps to wait to make sure that tuning is done (Default is 5).
        maxsteps (int): Number of maximum stepping-out steps (Default is 10^4).
        mu (float): Scale factor (Default value is 1.0), this will be tuned if tune=True.
        maxiter (int): Number of maximum Expansions/Contractions (Default is 10^4).
        pool (bool): External pool of workers to distribute workload to multiple CPUs (default is None).
        vectorize (bool): If true (default is False), logprob_fn receives not just one point but an array of points, and returns an array of likelihoods.
        blobs_dtype (list): List containing names and dtypes of blobs metadata e.g. ``[("log_prior", float), ("mean", float)]``. It's useful when you want to save multiple species of metadata. Default is None.
        verbose (bool): If True (default) print log statements.
        check_walkers (bool): If True (default) then check that ``nwalkers >= 2*ndim`` and even.
        shuffle_ensemble (bool): If True (default) then shuffle the ensemble of walkers in every iteration before splitting it.
        light_mode (bool): If True (default is False) then no expansions are performed after the tuning phase. This can significantly reduce the number of log likelihood evaluations but works best in target distributions that are apprroximately Gaussian.
    """
    def __init__(self,
                 nwalkers,
                 ndim,
                 logprob_fn,
                 args=None,
                 kwargs=None,
                 moves=None,
                 tune=True,
                 tolerance=0.05,
                 patience=5,
                 maxsteps=10000,
                 mu=1.0,
                 maxiter=10000,
                 pool=None,
                 vectorize=False,
                 blobs_dtype=None,
                 verbose=True,
                 check_walkers=True,
                 shuffle_ensemble=True,
                 light_mode=False):

        # Set up logger
        self.logger = logging.getLogger()
        for handler in self.logger.handlers[:]:
            self.logger.removeHandler(handler)
        handler = logging.StreamHandler()
        self.logger.addHandler(handler)
        if verbose:
            self.logger.setLevel(logging.INFO)
        else:
            self.logger.setLevel(logging.WARNING)

        # Parse the move schedule
        if moves is None:
            self._moves = [DifferentialMove()]
            self._weights = [1.0]
        elif isinstance(moves, Iterable):
            try:
                self._moves, self._weights = zip(*moves)
            except TypeError:
                self._moves = moves
                self._weights = np.ones(len(moves))
        else:
            self._moves = [moves]
            self._weights = [1.0]
        self._weights = np.atleast_1d(self._weights).astype(float)
        self._weights /= np.sum(self._weights)

        # Set up Log Probability
        self.logprob_fn = _FunctionWrapper(logprob_fn, args, kwargs)

        # Set up walkers
        self.nwalkers = int(nwalkers)
        self.ndim = int(ndim)
        self.check_walkers = check_walkers
        if self.check_walkers:
            if self.nwalkers < 2 * self.ndim:
                raise ValueError("Please provide at least (2 * ndim) walkers.")
            elif self.nwalkers % 2 == 1:
                raise ValueError("Please provide an even number of walkers.")
        self.shuffle_ensemble = shuffle_ensemble

        # Set up Slice parameters
        self.mu = mu
        self.mus = []
        self.mus.append(self.mu)
        self.tune = tune
        self.maxsteps = maxsteps
        self.patience = patience
        self.tolerance = tolerance
        self.nexps = []
        self.ncons =  []

        # Set up maximum number of Expansions/Contractions
        self.maxiter = maxiter

        # Set up pool of workers
        self.pool = pool
        self.vectorize = vectorize

        # Set up blobs dtype
        self.blobs_dtype = blobs_dtype

        # Initialise Saving space for samples
        self.samples = samples(self.ndim, self.nwalkers)

        # Initialise iteration counter and state
        self.iteration = 0
        self.state_X = None
        self.state_Z = None
        self.state_blobs = None

        # Light mode
        self.light_mode = light_mode


    def run(self, *args, **kwargs):
        logging.warning('The run method has been deprecated and it will be removed. Please use the new run_mcmc method.')
        return self.run_mcmc(*args, **kwargs)


    def reset(self):
        """
        Reset the state of the sampler. Delete any samples stored in memory.
        """
        self.samples = samples(self.ndim, self.nwalkers)
    

    def get_chain(self, flat=False, thin=1, discard=0):
        """
        Get the Markov chain containing the samples.

        Args:
            flat (bool) : If True then flatten the chain into a 2D array by combining all walkers (default is False).
            thin (int) : Thinning parameter (the default value is 1).
            discard (int) : Number of burn-in steps to be removed from each walker (default is 0). A float number between
                0.0 and 1.0 can be used to indicate what percentage of the chain to be discarded as burnin.

        Returns:
            Array object containg the Markov chain samples (2D if flat=True, 3D if flat=False).
        """

        if discard < 1.0:
            discard  = int(discard * np.shape(self.chain)[0])

        if flat:
            return self.samples.flatten(discard=discard, thin=thin)
        else:
            return self.chain[discard::thin,:,:]

    
    def get_log_prob(self, flat=False, thin=1, discard=0):
        """
        Get the value of the log probability function evalutated at the samples of the Markov chain.

        Args:
            flat (bool) : If True then flatten the chain into a 1D array by combining all walkers (default is False).
            thin (int) : Thinning parameter (the default value is 1).
            discard (int) : Number of burn-in steps to be removed from each walker (default is 0). A float number between
                0.0 and 1.0 can be used to indicate what percentage of the chain to be discarded as burnin.

        Returns:
            Array containing the value of the log probability at the samples of the Markov chain (1D if flat=True, 2D otherwise).
        """
        if discard < 1.0:
            discard  = int(discard * np.shape(self.chain)[0])

        if flat:
            return self.samples.flatten_logprob(discard=discard, thin=thin)
        else:
            return self.samples.logprob[discard::thin,:]


    def get_blobs(self, flat=False, thin=1, discard=0):
        """
        Get the values of the blobs at each step of the chain.

        Args:
            flat (bool) : If True then flatten the chain into a 1D array by combining all walkers (default is False).
            thin (int) : Thinning parameter (the default value is 1).
            discard (int) : Number of burn-in steps to be removed from each walker (default is 0). A float number between 0.0 and 1.0 can be used to indicate what percentage of the chain to be discarded as burnin.

        Returns:
            (structured) numpy array containing the values of the blobs at each step of the chain.
        """
        if discard < 1.0:
            discard  = int(discard * np.shape(self.chain)[0])

        if flat:
            return self.samples.flatten_blobs(discard=discard, thin=thin)
        else:
            return self.samples.blobs[discard::thin,:]


    @property
    def chain(self):
        """
        Returns the chains.

        Returns:
            Returns the chains of shape (nsteps, nwalkers, ndim).
        """
        return self.samples.chain


    @property
    def act(self):
        """
        Integrated Autocorrelation Time (IAT) of the Markov Chain.

        Returns:
            Array with the IAT of each parameter.
        """
        return AutoCorrTime(self.chain[int(self.nsteps/(self.thin*2.0)):,:,:])


    @property
    def ess(self):
        """
        Effective Sampling Size (ESS) of the Markov Chain.

        Returns:
            ESS
        """
        return self.nwalkers * self.samples.length / np.mean(self.act)


    @property
    def ncall(self):
        """
        Number of Log Prob calls.

        Returns:
            ncall
        """
        return np.sum(self.neval)


    @property
    def efficiency(self):
        """
        Effective Samples per Log Probability Evaluation.

        Returns:
            efficiency
        """
        return self.ess / self.ncall


    @property
    def scale_factor(self):
        """
        Scale factor values during tuning.

        Returns:
            scale factor mu
        """
        return np.asarray(self.mus)

    @property
    def get_last_sample(self):
        logging.warning('The ``get_last_sample`` property is deprecated and it will be removed in a future release.\n' + 'Please use the method ``get_last_sample()`` instead.')

        return self.chain[-1]

    
    def get_last_sample(self):
        """
        Return the last position of the walkers.
        """
        return self.chain[-1]

    
    def get_last_log_prob(self):
        """
        Return the log probability values for the last position of the walkers.
        """
        return self.samples.logprob[-1]

    
    def get_last_blobs(self):
        """
        Return the blobs for the last position of the walkers.
        """
        return self.samples.blobs[-1]


    @property
    def summary(self):
        """
        Summary of the MCMC run.
        """
        logging.info('Summary')
        logging.info('-------')
        logging.info('Number of Generations: ' + str(self.samples.length))
        logging.info('Number of Parameters: ' + str(self.ndim))
        logging.info('Number of Walkers: ' + str(self.nwalkers))
        logging.info('Number of Tuning Generations: ' + str(len(self.mus)))
        logging.info('Scale Factor: ' + str(round(self.mu,6)))
        logging.info('Mean Integrated Autocorrelation Time: ' + str(round(np.mean(self.act),2)))
        logging.info('Effective Sample Size: ' + str(round(self.ess,2)))
        logging.info('Number of Log Probability Evaluations: ' + str(self.ncall))
        logging.info('Effective Samples per Log Probability Evaluation: ' + str(round(self.efficiency,6)))
        if self.thin > 1:
            logging.info('Thinning rate: ' + str(self.thin))
    

    def compute_log_prob(self, coords):
        """
        Calculate the vector of log-probability for the walkers

        Args:
            coords: (ndarray[..., ndim]) The position vector in parameter space where the probability should be calculated.
        Returns:
            log_prob: A vector of log-probabilities with one entry for each walker in this sub-ensemble.
            blob: The list of meta data returned by the ``log_post_fn`` at this position or ``None`` if nothing was returned.
        """
        p = coords

        # Check that the parameters are in physical ranges.
        if np.any(np.isinf(p)):
            raise ValueError("At least one parameter value was infinite")
        if np.any(np.isnan(p)):
            raise ValueError("At least one parameter value was NaN")

        # Run the log-probability calculations (optionally in parallel).
        if self.vectorize:
            results = self.logprob_fn(p)
        else:
            results = list(self.distribute(self.logprob_fn, (p[i] for i in range(len(p)))))

        try:
            log_prob = np.array([float(l[0]) for l in results])
            blob = [l[1:] for l in results]
        except (IndexError, TypeError):
            log_prob = np.array([float(l) for l in results])
            blob = None
        else:
            # Get the blobs dtype
            if self.blobs_dtype is not None:
                dt = self.blobs_dtype
            else:
                try:
                    dt = np.atleast_1d(blob[0]).dtype
                except ValueError:
                    dt = np.dtype("object")
            blob = np.array(blob, dtype=dt)

            # Deal with single blobs properly
            shape = blob.shape[1:]
            if len(shape):
                axes = np.arange(len(shape))[np.array(shape) == 1] + 1
                if len(axes):
                    blob = np.squeeze(blob, tuple(axes))

        # Check for log_prob returning NaN.
        if np.any(np.isnan(log_prob)):
            raise ValueError("Probability function returned NaN")

        return log_prob, blob


    def run_mcmc(self,
                 start,
                 nsteps=1000,
                 thin=1,
                 progress=True,
                 log_prob0=None,
                 blobs0=None,
                 thin_by=1):
        '''
        Run MCMC.

        Args:
            start (float) : Starting point for the walkers. If ``None`` then the sampler proceeds
                from the last known position of the walkers.
            nsteps (int): Number of steps/generations (default is 1000).
            thin (float): Thin the chain by this number (default is 1, no thinning).
            progress (bool): If True (default), show progress bar.
            log_prob0 (float) : Log probability values of the walkers. Default is ``None``.
            blobs0 (float) : Blob value of the walkers. Default is ``None``.
            thin_by (float): If you only want to store and yield every
                ``thin_by`` samples in the chain, set ``thin_by`` to an
                integer greater than 1. When this is set, ``iterations *
                thin_by`` proposals will be made.
        '''
        
        for _ in self.sample(start,
                             log_prob0=log_prob0,
                             blobs0=blobs0,
                             iterations=nsteps,
                             thin=thin,
                             thin_by=thin_by,
                             progress=progress):
            pass
        

    def sample(self,
            start,
            log_prob0=None,
            blobs0=None,
            iterations=1,
            thin=1,
            thin_by=1,
            progress=True):
        '''
        Advance the chain as a generator. The current iteration index of the generator is given by the ``sampler.iteration`` property.

        Args:
            start (float) : Starting point for the walkers.
            log_prob0 (float) : Log probability values of the walkers. Default is ``None``.
            blobs0 (float) : Blob value of the walkers. Default is ``None``.
            iterations (int): Number of steps to generate (default is 1).
            thin (float): Thin the chain by this number (default is 1, no thinning).
            thin_by (float): If you only want to store and yield every
                ``thin_by`` samples in the chain, set ``thin_by`` to an
                integer greater than 1. When this is set, ``iterations *
                thin_by`` proposals will be made.
            progress (bool): If True (default), show progress bar.
        '''
        # Define task distributer
        if self.pool is None:
            self.distribute = map
        else:
            self.distribute = self.pool.map

        # Initialise ensemble of walkers
        logging.info('Initialising ensemble of %d walkers...', self.nwalkers)
        if start is not None:
            if np.shape(start) != (self.nwalkers, self.ndim):
                raise ValueError('Incompatible input dimensions! \n' +
                                 'Please provide array of shape (nwalkers, ndim) as the starting position.')
            X = np.copy(start)
            if log_prob0 is None:
                Z, blobs = self.compute_log_prob(X)
            else:
                Z = np.copy(log_prob0)
                blobs = blobs0
        elif (self.state_X is not None) and (self.state_Z is not None):
            X = np.copy(self.state_X)
            Z = np.copy(self.state_Z)
            blobs = self.state_blobs
        else:
            raise ValueError("Cannot have `start=None` if run_mcmc has never been called before.")


        if not np.all(np.isfinite(Z)):
            raise ValueError('Invalid walker initial positions! \n' +
                             'Initialise walkers from positions of finite log probability.')
        batch = list(np.arange(self.nwalkers))

        # Extend saving space
        self.thin = int(thin)
        self.thin_by = int(thin_by)
        
        if self.thin_by < 0:
            raise ValueError('Invalid `thin_by` argument.')
        elif self.thin < 0:
            raise ValueError('Invalid `thin` argument.')
        elif self.thin > 1 and self.thin_by == 1:
            self.nsteps = int(iterations)
            self.samples.extend(self.nsteps//self.thin, blobs)
            self.ncheckpoint = self.thin
        elif self.thin_by > 1 and self.thin == 1:
            self.nsteps = int(iterations*self.thin_by)
            self.samples.extend(self.nsteps//self.thin_by, blobs)
            self.ncheckpoint = self.thin_by
        elif self.thin == 1 and self.thin_by == 1:
            self.nsteps = int(iterations)
            self.samples.extend(self.nsteps, blobs)
            self.ncheckpoint = 1
        else:
            raise ValueError('Only one of `thin` and `thin_by` arguments can be used.')
        

        # Define Number of Log Prob Evaluations vector
        self.neval = np.zeros(self.nsteps, dtype=int)

        # Define tuning count
        ncount = 0

        # Initialise progress bar
        if progress:
            t = tqdm(total=self.nsteps, desc='Sampling progress : ')

        # Main Loop
        for i in range(self.nsteps):

            # Initialise number of expansions & contractions
            nexp = 0
            ncon = 0

            move = np.random.choice(self._moves, p=self._weights)

            # Shuffle and split ensemble
            if self.shuffle_ensemble:
                np.random.shuffle(batch)
            batch0 = batch[:int(self.nwalkers/2)]
            batch1 = batch[int(self.nwalkers/2):]
            sets = [[batch0,batch1],[batch1,batch0]]

            # Loop over two sets
            for ensembles in sets:
                indeces = np.arange(int(self.nwalkers/2))
                # Define active-inactive ensembles
                active, inactive = ensembles

                # Compute directions
                directions, tune_once = move.get_direction(X[inactive], self.mu)

                # Get Z0 = LogP(x0)
                Z0 = Z[active] - np.random.exponential(size=int(self.nwalkers/2))

                # Set Initial Interval Boundaries
                L = - np.random.uniform(0.0,1.0,size=int(self.nwalkers/2))
                R = L + 1.0

                # Parallel stepping-out
                J = np.floor(self.maxsteps * np.random.uniform(0.0,1.0,size=int(self.nwalkers/2)))
                K = (self.maxsteps - 1) - J

                # Initialise number of Log prob calls
                ncall = 0

                # Left stepping-out initialisation
                mask_J = np.full(int(self.nwalkers/2),True)
                Z_L = np.empty(int(self.nwalkers/2))
                X_L = np.empty((int(self.nwalkers/2),self.ndim))

                # Right stepping-out initialisation
                mask_K = np.full(int(self.nwalkers/2),True)
                Z_R = np.empty(int(self.nwalkers/2))
                X_R = np.empty((int(self.nwalkers/2),self.ndim))

                cnt = 0
                # Stepping-Out procedure
                while len(mask_J[mask_J])>0 or len(mask_K[mask_K])>0:
                    if len(mask_J[mask_J])>0:
                        cnt += 1
                    if len(mask_K[mask_K])>0:
                        cnt += 1
                    if cnt > self.maxiter:
                        raise RuntimeError('Number of expansions exceeded maximum limit! \n' +
                                           'Make sure that the pdf is well-defined. \n' +
                                           'Otherwise increase the maximum limit (maxiter=10^4 by default).')

                    for j in indeces[mask_J]:
                        if J[j] < 1:
                            mask_J[j] = False

                    for j in indeces[mask_K]:
                        if K[j] < 1:
                            mask_K[j] = False

                    X_L[mask_J] = directions[mask_J] * L[mask_J][:,np.newaxis] + X[active][mask_J]
                    X_R[mask_K] = directions[mask_K] * R[mask_K][:,np.newaxis] + X[active][mask_K]

                    if len(X_L[mask_J]) + len(X_R[mask_K]) < 1:
                        Z_L[mask_J] = np.array([])
                        Z_R[mask_K] = np.array([])
                        cnt -= 1
                    else:
                        Z_LR_masked, _ = self.compute_log_prob(np.concatenate([X_L[mask_J],X_R[mask_K]]))
                        #Z_LR_masked = np.array(list(self.distribute(self.logprob_fn, np.concatenate([X_L[mask_J],X_R[mask_K]]))))
                        Z_L[mask_J] = Z_LR_masked[:X_L[mask_J].shape[0]]
                        Z_R[mask_K] = Z_LR_masked[X_L[mask_J].shape[0]:]

                    for j in indeces[mask_J]:
                        ncall += 1
                        if Z0[j] < Z_L[j]:
                            L[j] = L[j] - 1.0
                            J[j] = J[j] - 1
                            nexp += 1
                        else:
                            mask_J[j] = False

                    for j in indeces[mask_K]:
                        ncall += 1
                        if Z0[j] < Z_R[j]:
                            R[j] = R[j] + 1.0
                            K[j] = K[j] - 1
                            nexp += 1
                        else:
                            mask_K[j] = False


                # Shrinking procedure
                Widths = np.empty(int(self.nwalkers/2))
                Z_prime = np.empty(int(self.nwalkers/2))
                X_prime = np.empty((int(self.nwalkers/2),self.ndim))
                if blobs is not None:
                    blobs_prime = np.empty(int(self.nwalkers/2), dtype=np.dtype((blobs[0].dtype, blobs[0].shape)))
                mask = np.full(int(self.nwalkers/2),True)

                cnt = 0
                while len(mask[mask])>0:
                    # Update Widths of intervals
                    Widths[mask] = L[mask] + np.random.uniform(0.0,1.0,size=len(mask[mask])) * (R[mask] - L[mask])

                    # Compute New Positions
                    X_prime[mask] = directions[mask] * Widths[mask][:,np.newaxis] + X[active][mask]
                    

                    # Calculate LogP of New Positions
                    if blobs is None:
                        Z_prime[mask], _ = self.compute_log_prob(X_prime[mask])
                        #Z_prime[mask] = np.array(list(self.distribute(self.logprob_fn, X_prime[mask])))
                    else:
                        Z_prime[mask], blobs_prime[mask] = self.compute_log_prob(X_prime[mask])

                    # Count LogProb calls
                    ncall += len(mask[mask])

                    # Shrink slices
                    for j in indeces[mask]:
                        if Z0[j] < Z_prime[j]:
                            mask[j] = False
                        else:
                            if Widths[j] < 0.0:
                                L[j] = Widths[j]
                                ncon += 1
                            elif Widths[j] > 0.0:
                                R[j] = Widths[j]
                                ncon += 1

                    cnt += 1
                    if cnt > self.maxiter:
                        raise RuntimeError('Number of contractions exceeded maximum limit! \n' +
                                           'Make sure that the pdf is well-defined. \n' +
                                           'Otherwise increase the maximum limit (maxiter=10^4 by default).')

                # Update Positions
                X[active] = X_prime
                Z[active] = Z_prime
                if blobs is not None:
                    blobs[active] = blobs_prime
                self.neval[i] += ncall

            # Tune scale factor using Robbins-Monro optimization
            if self.tune and tune_once:
                self.nexps.append(nexp)
                self.ncons.append(ncon)
                nexp = max(1, nexp) # This is to prevent the optimizer from getting stuck
                self.mu *= 2.0 * nexp / (nexp + ncon)
                self.mus.append(self.mu)
                if np.abs(nexp / (nexp + ncon) - 0.5) < self.tolerance:
                    ncount += 1
                if ncount > self.patience:
                    self.tune = False
                    if self.light_mode:
                        self.mu *= (1.0 + nexp/self.nwalkers)
                        self.maxsteps = 1

            # Save samples
            if (i+1) % self.ncheckpoint == 0:
                self.samples.save(X, Z, blobs)

            # Update progress bar
            if progress:
                t.update()

            # Update iteration counter and state variables
            self.iteration = i + 1
            self.state_X = np.copy(X)
            self.state_Z = np.copy(Z)
            self.state_blobs = blobs

            # Yield current state
            if (i+1) % self.ncheckpoint == 0:
                yield (X, Z, blobs)

        # Close progress bar
        if progress:
            t.close()


class sampler(EnsembleSampler):
    def __init__(self, *args, **kwargs):
        logging.warning('The sampler class has been deprecated. Please use the new EnsembleSampler class.')
        super().__init__(*args, **kwargs)


import numpy as np

class _FunctionWrapper(object):
    """
    This is a hack to make the likelihood function pickleable when ``args``
    or ``kwargs`` are also included.

    Args:
        f (callable) : Log Probability function.
        args (list): Extra arguments to be passed into the logprob.
        kwargs (dict): Extra arguments to be passed into the logprob.

    Returns:
        Log Probability function.
    """

    def __init__(self, f, args, kwargs):
        self.f = f
        self.args = [] if args is None else args
        self.kwargs = {} if kwargs is None else kwargs

    def __call__(self, x):
        try:
            return self.f(x, *self.args, **self.kwargs)
        except:  # pragma: no cover
            import traceback

            print("zeus: Exception while calling your likelihood function:")
            print("  params:", x)
            print("  args:", self.args)
            print("  kwargs:", self.kwargs)
            print("  exception:")
            traceback.print_exc()
            raise



__bibtex__ = """
@article{karamanis2021zeus,
             title={zeus: A Python implementation of Ensemble Slice Sampling for efficient Bayesian parameter inference},
             author={Karamanis, Minas and Beutler, Florian and Peacock, John A},
             journal={arXiv preprint arXiv:2105.03468},
             year={2021}
            }

    @article{karamanis2020ensemble,
             title = {Ensemble slice sampling: Parallel, black-box and gradient-free inference for correlated & multimodal distributions},
             author = {Karamanis, Minas and Beutler, Florian},
             journal = {arXiv preprint arXiv: 2002.06212},
             year = {2020}
            }
"""

__version__ = "2.3.1"
__url__ = "https://zeus-mcmc.readthedocs.io"
__author__ = "Minas Karamanis"
__email__ = "minaskar@gmail.com"
__license__ = "GPL-3.0"
__description__ = "Lightning Fast MCMC"


from .ensemble import *
from .parallel import ChainManager
from .autocorr import AutoCorrTime
from .plotting import cornerplot


import numpy as np

class samples:
    '''
    Creates object that stores the samples.
    Args:
        ndim (int): Number of dimensions/paramters
        nwalkers (int): Number of walkers.

    '''

    def __init__(self, ndim, nwalkers):
        self.initialised = False
        self.index = 0
        self.ndim = ndim
        self.nwalkers = nwalkers


    def extend(self, n, blobs):
        """
        Method to extend saving space.
        Args:
            n (int) : Extend space by n slots.
        """
        if self.initialised:
            ext = np.empty((n,self.nwalkers,self.ndim))
            self.samples = np.concatenate((self.samples,ext),axis=0)
            ext = np.empty((n,self.nwalkers))
            self.logp = np.concatenate((self.logp,ext),axis=0)
            if blobs is not None:
                dt = np.dtype((blobs[0].dtype, blobs[0].shape))
                ext = np.empty((n,self.nwalkers), dtype=dt)
                self.blobs = np.concatenate((self.blobs, ext),axis=0)
        else:
            self.samples = np.empty((n,self.nwalkers,self.ndim))
            self.logp = np.empty((n,self.nwalkers))
            if blobs is not None:
                dt = np.dtype((blobs[0].dtype, blobs[0].shape))
                self.blobs = np.empty((n,self.nwalkers), dtype=dt)
            self.initialised = True


    def save(self, x, logp, blobs):
        """
        Save sample into the storage.
        Args:
            x (ndarray): Samples to be appended into the storage.
            logp (ndarray): Logprob values to be appended into the storage.
        """
        self.samples[self.index] = x
        self.logp[self.index] = logp
        if blobs is not None:
            self.blobs[self.index] = blobs
        self.index += 1


    @property
    def chain(self):
        """
        Chain property.

        Returns:
            3D array of shape (nsteps,nwalkers,ndim) containing the samples.
        """
        return self.samples


    @property
    def length(self):
        """
        Number of samples per walker.

        Returns:
            The total number of samples per walker.
        """
        length, _, _ = np.shape(self.chain)
        return length


    def flatten(self, discard=0, thin=1):
        """
        Flatten samples by thinning them, removing the burn in phase, and combining all the walkers.

        Args:
            discard (int): Number of burn-in steps to be removed from each walker (default is 0).
            thin (int): Thinning parameter (the default value is 1).

        Returns:
            2D object containing the ndim flattened chains.
        """
        return self.chain[discard::thin,:,:].reshape((-1,self.ndim), order='F')


    @property
    def logprob(self):
        """
        Chain property.

        Returns:
            2D array of shape (nwalkers,nsteps) containing the log-probabilities.
        """
        return self.logp


    def flatten_logprob(self, discard=0, thin=1):
        """
        Flatten log probability by thinning the chain, removing the burn in phase, and combining all the walkers.

        Args:
            discard (int): Number of burn-in steps to be removed from each walker (default is 0).
            thin (int): Thinning parameter (the default value is 1).

        Returns:
            1D object containing the logprob of the flattened chains.
        """
        return self.logprob[discard::thin,:].reshape((-1,), order='F')


    def flatten_blobs(self, discard=0, thin=1):
        """
        Flatten blobs by thinning the chain, removing the burn in phase, and combining all the walkers.

        Args:
            discard (int): Number of burn-in steps to be removed from each walker (default is 0).
            thin (int): Thinning parameter (the default value is 1).

        Returns:
            (structured) NumPy array containing the blobs metadata.
        """
        return self.blobs[discard::thin,:].reshape((-1,), order='F')


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", 'r') as dependencies:
    requirements = [pkg.strip() for pkg in dependencies]

setuptools.setup(
    name="zeus-mcmc",
    version="2.3.1",
    author="Minas Karamanis",
    author_email="minaskar@gmail.com",
    description="zeus: Lightning Fast MCMC",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/minaskar/zeus",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
    ],
    install_requires=requirements,
    python_requires='>=3.6',
)
